﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {
  [StorableClass]
  [Item("GenerationalDistanceAnalyzer", "Computes the enclosed Hypervolume between the current front and a given reference Point")]
  public class HypervolumeAnalyzer : MOTFAnalyzer {
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
    }
    [StorableConstructor]
    protected HypervolumeAnalyzer(bool deserializing) : base(deserializing) { }
    public HypervolumeAnalyzer(HypervolumeAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new HypervolumeAnalyzer(this, cloner);
    }

    public IValueParameter<DoubleArray> ReferencePointParameter {
      get {
        return (IValueParameter<DoubleArray>)Parameters["ReferencePoint"];
      }
    }

    public IValueParameter<DoubleValue> BestKnownHyperVolumeParameter {
      get {
        return (IValueParameter<DoubleValue>)Parameters["BestKnownHyperVolume"];
      }
      set {
        Parameters["BestKnownHyperVolume"].ActualValue = value;
      }
    }

    public HypervolumeAnalyzer() {
      Parameters.Add(new ValueParameter<DoubleArray>("ReferencePoint", "The reference point for hypervolume calculation"));
      Parameters.Add(new ValueParameter<DoubleValue>("BestKnownHyperVolume", "The currently best known hypervolume"));
    }

    private void RegisterEventHandlers() {
      ReferencePointParameter.ValueChanged += ReferencePointParameterOnValueChanged;
    }

    private void ReferencePointParameterOnValueChanged(object sender, EventArgs e) {
      BestKnownHyperVolumeParameter.Value = new DoubleValue(0);
    }

    public override void Analyze(Individual[] individuals, double[][] qualities, ResultCollection results) {
      if (qualities == null || qualities.Length < 1) return;
      int objectives = qualities[0].Length;

      if (!results.ContainsKey("Hypervolume")) results.Add(new Result("Hypervolume", typeof(DoubleValue)));
      if (!results.ContainsKey("Absolute Distance to BestKnownHypervolume")) results.Add(new Result("Absolute Distance to BestKnownHypervolume", typeof(DoubleValue)));
      IEnumerable<double[]> front = NonDominatedSelect.selectNonDominatedVectors(qualities, TestFunctionParameter.ActualValue.Maximization(objectives), true);
      double hv = front.Any() ? Hypervolume.Calculate(front, TestFunctionParameter.ActualValue.ReferencePoint(objectives), TestFunctionParameter.ActualValue.Maximization(objectives)) : 0;
      double best;
      if (!results.ContainsKey("BestKnownHypervolume")) {
        results.Add(new Result("BestKnownHypervolume", typeof(DoubleValue)));
        best = BestKnownHyperVolumeParameter.Value.Value;
      } else {
        best = ((DoubleValue)(results["BestKnownHypervolume"].Value)).Value;
      }
      if (Double.IsNaN(best)) best = hv; else best = Math.Max(best, hv);

      double diff = best - hv;
      if (diff == 0) {
        BestKnownFrontParameter.ActualValue = new DoubleMatrix(MultiObjectiveTestFunctionProblem.To2D(qualities));
      }

      results["Hypervolume"].Value = new DoubleValue(hv);
      results["Absolute Distance to BestKnownHypervolume"].Value = new DoubleValue(diff);
      results["BestKnownHypervolume"].Value = new DoubleValue(best);

    }
  }
}
